#ifndef CoCoA_MachineInteger_H
#define CoCoA_MachineInteger_H

//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/assert.H"

#include <limits>
//using std::numeric_limits;
#include <iosfwd>
//using std::ostream;

namespace CoCoA
{

  // This class is deliberately inline.
  class MachineInteger
  {
  public:
    MachineInteger(signed char n): myValue(n), IamNegative(n<0) {}
    MachineInteger(unsigned char n): myValue(n), IamNegative(false) {}
    MachineInteger(short n): myValue(n), IamNegative(n<0) {}
    MachineInteger(unsigned short n): myValue(n), IamNegative(false) {}
    MachineInteger(int n): myValue(n), IamNegative(n<0) {}
    MachineInteger(unsigned int n): myValue(n), IamNegative(false) {}
    MachineInteger(long n): myValue(n), IamNegative(n<0) {}
    MachineInteger(unsigned long n): myValue(n), IamNegative(false) {}
    friend bool IsNegative(const MachineInteger& n);
    friend unsigned long AsUnsignedLong(const MachineInteger& n);
    friend long AsSignedLong(const MachineInteger& n);
    friend unsigned long abs(const MachineInteger& n);
  private:
    unsigned long myValue;
    bool IamNegative;
  };

  inline  bool IsNegative(const MachineInteger& n)
  { return n.IamNegative; }

  inline unsigned long AsUnsignedLong(const MachineInteger& n)
  {
    CoCoA_ASSERT(!n.IamNegative);
    return n.myValue;
  }

  inline long AsSignedLong(const MachineInteger& n)
  {
    CoCoA_ASSERT(n.IamNegative || n.myValue <= static_cast<unsigned long>(std::numeric_limits<long>::max()));
    return n.myValue;
  }

  inline unsigned long abs(const MachineInteger& n)
  { if (n.IamNegative) return -n.myValue; else return n.myValue; }

  std::ostream& operator<<(std::ostream& out, const MachineInteger& n);

  inline bool IsZero(const MachineInteger& n)
  { return AsUnsignedLong(n) == 0; }

//  Will these two fns ever be useful???
//   inline bool IsOne(const MachineInteger& n)
//   { return (!IsNegative(n) && AsUnsignedLong(n) == 1); }

//   inline bool IsMinusOne(const MachineInteger& n)
//   { return (IsNegative(n) && AsSignedLong(n) == -1); }

  inline int sign(const MachineInteger& n)
  { if (IsNegative(n)) return -1; if (IsZero(n)) return 0; return 1; }

  inline bool IsOdd(const MachineInteger& n)
  { return abs(n)&1; }

  inline bool IsEven(const MachineInteger& n)
  { return !IsOdd(n); }


  int cmp(const MachineInteger& n1, const MachineInteger& n2);
  unsigned long gcd(const MachineInteger& a, const MachineInteger& b);


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/MachineInteger.H,v 1.2 2007/10/30 17:14:12 abbott Exp $
// $Log: MachineInteger.H,v $
// Revision 1.2  2007/10/30 17:14:12  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
//

#endif
