//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for classes matrix and MatrixBase

#include "CoCoA/matrix.H"
#include "CoCoA/assert.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/OpenMath.H"

#include <iostream>
using std::ostream;
//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;

namespace CoCoA
{

//   RingElem MatrixVector::operator[] (size_t i)
//   {
//     if (i >= myCoords.myNumCompts())
//       CoCoA_ERROR("MatrixVector index too large", "MatrixVector[i]")
//     RingElem ans(myR);
//     myEntry(ans, i);
//     return ans;
//   }


//   void set(MatrixVector& v, size_t i, ConstRefRingElem r)
//   {
//     if (owner(r) != owner(v))
//       CoCoA_ERROR("Incompatible rings", "set(MatrixVector,i,r)");
//     if (i >= myCoords.myNumCompts())
//       CoCoA_ERROR("MatrixVector index too large", "set(MatrixVector,i,r)")
//     myCoords.mySetEntry(i, r);
//   }



  ConstMatrixBase::ConstMatrixBase():
      myRefCount(0)
  {}


  ConstMatrixBase::~ConstMatrixBase()
  {}  // does nothing but has to exist for C++ to compile


  MatrixBase::MatrixBase()
  {}  // ctor for ConstMatrixBase zeroes myRefCount


  MatrixBase::~MatrixBase()
  {}  // does nothing but has to exist for C++ to compile

  void ConstMatrixBase::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    vector<RingElem> ans(myNumCols(), zero(myBaseRing()));
    for (size_t j=0; j < myNumCols(); ++j)
      for (size_t i=0; i < myNumRows(); ++i)
        ans[j] += v[i]*myEntry(i, j);
    // We have successfully computed the answer, now swap it in.
    swap(lhs, ans); // does this invalidate iterators on lhs?
//     for (size_t j=0; j < myNumCols(); ++j)
//       swap(lhs[j], ans[j]);
  }


  void ConstMatrixBase::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    vector<RingElem> ans(myNumRows(), zero(myBaseRing()));
    for (size_t i=0; i < myNumRows(); ++i)
      for (size_t j=0; j < myNumCols(); ++j)
        ans[i] += myEntry(i, j)*v[j];

    // We have successfully computed the answer, now swap it in.
    swap(lhs, ans); // does this invalidate iterators on lhs?
//     for (size_t i=0; i < myNumRows(); ++i)
//       swap(lhs[i], ans[i]);
  }


  bool ConstMatrixBase::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    for (size_t j=0; j < myNumCols(); ++j)
      if (!IsZero(myEntry(i,j))) return false;
    return true;
  }


  bool ConstMatrixBase::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    for (size_t i=0; i < myNumRows(); ++i)
      if (!IsZero(myEntry(i,j))) return false;
    return true;
  }


  void ConstMatrixBase::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumCols());
    DetByGauss(d, ConstMatrix(this));
  }


  size_t ConstMatrixBase::myRank() const
  {
    return RankByGauss(ConstMatrix(this));
  }


  void ConstMatrixBase::myOutputSelf(std::ostream& out) const
  {
    out << "matrix(" << myBaseRing() << ")(" << myNumRows() << ", " << myNumCols() << ")\n[\n";
    for (size_t i=0; i < myNumRows(); ++i)
    {
      out << " [";
      for (size_t j=0; j < myNumCols(); ++j)
      {
        if (j > 0) out << ", ";
        out << myEntry(i, j);
      }
      out << "]";
      if (i != myNumRows()-1) out << ",";
      out << "\n";
    }
    out << "]\n";
  }


  std::ostream& operator<<(std::ostream& out, ConstMatrix M)
  {
    M->myOutputSelf(out);
    return out;
  }


  void ConstMatrixBase::myOutputSelf(OpenMathOutput& OMOut) const
  {
    OMOut->mySendApplyStart();
    OMOut << OpenMathSymbol("linalg2", "matrix");
    OMOut << myBaseRing() << myNumRows() << myNumCols();
    for (size_t i=0; i < myNumRows(); ++i)
    {
      OMOut->mySendApplyStart();
      OMOut << OpenMathSymbol("linalg2", "matrixrow"); //??? << myNumCols; ???
      for (size_t j=0; j < myNumCols(); ++j)
      {
        OMOut << myEntry(i, j); //??? print only raw value???
      }
      OMOut->mySendApplyEnd();
    }
    OMOut->mySendApplyEnd();
  }


  void ConstMatrixBase::myCheckIndices(std::size_t i, std::size_t j, const char* where) const
  {
    if (i >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, where);
    if (j >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, where);
  }


  ConstRefRingElem ConstMatrix::operator()(std::size_t i, std::size_t j) const
  {
    myPtr->myCheckIndices(i, j, "M(i,j)");
    return myPtr->myEntry(i, j);
  }


  void AssignZero(matrix& M)
  {
    M->myAssignZero();
  }


  void SetEntry(matrix& M, std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    if (owner(r) != BaseRing(M))
      CoCoA_ERROR("Incompatible rings", "SetEntry(M,i,j,r)");
    M->myCheckIndices(i, j, "SetEntry(M,i,j,r)");
    M->mySetEntry(i, j, r);
  }


  void SetEntry(matrix& M, std::size_t i, std::size_t j, long n)
  {
    M->myCheckIndices(i, j, "SetEntry(M,i,j,n)");
    M->mySetEntry(i, j, n);
  }


  void SetEntry(matrix& M, std::size_t i, std::size_t j, const ZZ& N)
  {
    M->myCheckIndices(i, j, "SetEntry(M,i,j,N)");
    M->mySetEntry(i, j, N);
  }


  bool IsZeroRow(ConstMatrix M, std::size_t i)
  {
    if (i >= M->myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "IsZeroRow");
    return M->myIsZeroRow(i);
  }


  bool IsZeroCol(ConstMatrix M, std::size_t j)
  {
    if (j >= M->myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "IsZeroCol");
    return M->myIsZeroCol(j);
  }


  RingElem det(ConstMatrix M)
  {
    if (NumRows(M) != NumRows(M))
      CoCoA_ERROR(ERR::NotSquareMatrix, "det");
    RingElem d(BaseRing(M));
    M->myDet(d);
    return d;
  }


  size_t rank(ConstMatrix M)
  {
    if (!IsIntegralDomain(BaseRing(M)))
      CoCoA_ERROR(ERR::NotIntegralDomain, "rank");
    return M->myRank();
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/matrix.C,v 1.2 2007/10/30 17:14:06 abbott Exp $
// $Log: matrix.C,v $
// Revision 1.2  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.4  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.3  2006/11/02 13:25:43  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.3  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.2  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.11  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.10  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.9  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.8  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.7  2005/03/31 16:59:42  cocoa
// Made special matrix ctors private, so a user has to pass via the
// pseudo-ctors (which do all the arg sanity checking).
//
// Revision 1.6  2005/03/30 17:15:14  cocoa
// Cleaned the SpecialMatrix code; a simple test compiles and
// runs fine.
//
// Revision 1.5  2005/03/29 17:19:59  cocoa
// -- added: rank
//
// Revision 1.4  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.3  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.9  2004/11/29 16:19:02  cocoa
// -- changed syntax for function det
//
// Revision 1.8  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.7  2004/11/11 13:55:10  cocoa
// -- minor changes for doxygen
//
// Revision 1.6  2004/11/02 14:49:03  cocoa
// -- new code for matrix orderings
//
// Revision 1.5  2004/07/20 15:04:06  cocoa
// The next step in the new "ring element" conversion process:
// handling the case of creating a "const RefRingElem" object
// (since C++ refuses to do this properly itself).
//
// Revision 1.4  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.3  2004/03/20 17:46:10  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.2  2004/03/09 16:15:41  cocoa
// First version of matrices.  A simple exmaple compiles and runs,
// so I'm checking in.
//
// Revision 1.1  2004/01/28 16:29:17  cocoa
// First attempt at implementing matrices.
//
//
